%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ReSaDE
% Author:           Mengyuan Zhang
% Created Date:     June 20, 2022

function [Fit_and_p, best_x, fitMaxVector, yhistory] = ...
    ReSaDE(deParameters, caseStudyData, otherParameters, low_habitat_limit, up_habitat_limit, iRuns)
fprintf('=================== Start trial %d ===================\n', iRuns);
%% Initialization
I_NP         = deParameters.I_NP;
F_weight     = deParameters.F_weight;
F_CR         = deParameters.F_CR;
I_D          = numel(up_habitat_limit); %Total Number of Variables
deParameters.nVariables=I_D;
FVr_minbound = low_habitat_limit;
FVr_maxbound = up_habitat_limit;
%fitness function
fnc=  otherParameters.fnc;

% initialize results recording
yhistory = [];

% For simpleDE
%mutation strategy
I_strategy   = deParameters.I_strategy; 
%Repair boundary method employed
BRM=deParameters.I_bnd_constr; %1: bring the value to bound violated
                               %2: repair in the allowed range

% pre-allocation of loop variables
I_itermax = 405;
fitMaxVector = nan(1,I_itermax);

% init generation counter
gen = 0; %iterations

rand('state',sum(otherParameters.iRuns)) %Guarantee same initial population


%% variable grouping
fprintf('---------- Start grouping ----------\n');
group = cell(1,4);
subidx1 = []; subidx2 = []; subidx3 = []; subidx4 = [];
for k = 1:I_D
    if FVr_minbound(k) == FVr_maxbound(k)
        subidx4 = [subidx4, k]; % dummy
    elseif FVr_minbound(k) == -FVr_maxbound(k)
        if FVr_maxbound(k) < 5
            subidx1 = [subidx1, k];
        else
            subidx2 = [subidx2, k]; 
        end
    else
        subidx3 = [subidx3, k]; 
    end
end

subidx1 = {subidx1};
group(1) = subidx1;
subidx2 = {subidx2};
group(2) = subidx2;
subidx3 = {subidx3};
group(3) = subidx3;
subidx4 = {subidx4};
group(4) = subidx4;

%% Group-wise optimization
% initialize platform
I_NP = 10;
No_solutions = 10;

otherParameters =setOtherParameters(caseStudyData, No_solutions, 1);
otherParameters.iRuns=iRuns; 
minPositionsMatrix=repmat(FVr_minbound,I_NP,1);
maxPositionsMatrix=repmat(FVr_maxbound,I_NP,1);
deParameters.minPositionsMatrix=minPositionsMatrix;
deParameters.maxPositionsMatrix=maxPositionsMatrix;

tmp_bestval = [];
tmp_bestx = [];
for j = 1:3
    %% run resade for each group %%%
    subLbound = minPositionsMatrix(:, group{j}); 
    subUbound = maxPositionsMatrix(:, group{j});
    I_D = length(group{j});
    % initial pop
    FM_pop=genpop(I_NP,I_D,subLbound,subUbound);
    FM_pop(1,:) = FVr_minbound(group{j});
    lb_pop = minPositionsMatrix;
    lb_pop(:,group{j}) =  FM_pop;  
    % evaluate the initial pop
    [S_val, ~]=feval(fnc,lb_pop,caseStudyData,otherParameters);
    yhistory = [yhistory; S_val];  
    [S_bestval,I_best_index] = min(S_val); % best individual correspond to the best worst performance
    FVr_bestmemit = FM_pop(I_best_index,:); % best member of current iteration
    best_x = lb_pop(I_best_index,:);

    % record the fitness value
    gen = gen +1;
    fitMaxVector(1,gen) = S_bestval;
    
    % parameter initialization
    FEs = 0;
%     ccm = deParameters.ccm_init;
    ccm = 0.35;
    sansde_iter = 34;
    linkp = deParameters.link_p;
    Cycle = 0;
    iter = 1;
    popsize = 10;
    bestval = S_bestval;
    display = 1;
    grouping_verbose = 0;
    
    Max_FEs = sansde_iter*I_NP;
    while (FEs < Max_FEs)
        Cycle = Cycle + 1;
        dorestart = 0; % run resade without restart
        if (FEs + (sansde_iter * popsize) > Max_FEs)
            sansde_iter = ceil((Max_FEs - FEs) / popsize);
            dorestart = 0;
        end
        [subpopnew, bestmemnew, bestvalnew, tracerst, ccm, used_FEs, fitMaxVector, gen, yhistory, ~] = sansde_restart(fnc, group{j}, FM_pop, S_val, best_x, bestval, subLbound, subUbound, sansde_iter, ccm, linkp, caseStudyData, otherParameters, fitMaxVector, gen, yhistory, dorestart, grouping_verbose);
        FEs = FEs + used_FEs;
        FM_pop = subpopnew;
        if bestvalnew < bestval
            best_x = bestmemnew;
            FVr_bestmemit = bestmemnew(group{j});
            bestval = bestvalnew;
        end
        iter = iter + 1;
    
        if(display == 1)
            fprintf(1, 'Cycle = %d, bestval = %e,\n', Cycle, bestval);
        end
        
        if(FEs >= Max_FEs)
            break;
        end
       
    end
    % record the current best for each group
    tmp_bestval = [tmp_bestval bestval];
    tmp_bestx = [tmp_bestx; best_x];
end
% sort the groups according to the current best
[groupval, groupidx] = sort(tmp_bestval, 'ascend');
% select the most effective group of variables
subidx = group{groupidx(1)};

fprintf('------ Finish variable selection ------\n');

%% %%%%%%%%%%%%%%%%%%% Optimization over selected group %%%%%%%%%%%%%%%%%%
fprintf('\n------------ Start ReSaDE ------------\n');
%% Warm-start using Simple DE
fprintf('******* Warm-start *******\n');
warm_verbose = 0;
% initialize platform
I_NP = 10;
No_solutions = 10;
otherParameters =setOtherParameters(caseStudyData, No_solutions, 1);
otherParameters.iRuns=iRuns; 
minPositionsMatrix=repmat(FVr_minbound,I_NP,1);
maxPositionsMatrix=repmat(FVr_maxbound,I_NP,1);
deParameters.minPositionsMatrix=minPositionsMatrix;
deParameters.maxPositionsMatrix=maxPositionsMatrix;
subLbound = minPositionsMatrix(:, subidx); 
subUbound = maxPositionsMatrix(:, subidx);

% initial pop
I_D = length(subidx);
FM_pop=genpop(I_NP,I_D,subLbound,subUbound);

FM_pop(1,:) = FVr_minbound(subidx);
curr_bestx = tmp_bestx(groupidx(1),:);
FM_pop(2,:) = curr_bestx(subidx);

lb_pop = minPositionsMatrix;
lb_pop(:,subidx) =  FM_pop;

bestmemit_list = [];

% evaluate the initial pop
[S_val, ~]=feval(fnc,lb_pop,caseStudyData,otherParameters);
yhistory = [yhistory; S_val];

[S_bestval,I_best_index] = min(S_val); 
FVr_bestmemit = FM_pop(I_best_index,:); 
bestmemit_list = [bestmemit_list, FVr_bestmemit];
best_x = lb_pop(I_best_index,:);

% record the fitness value
gen = gen +1;
fitMaxVector(1,gen) = S_bestval;

% simple DE
FVr_rot  = (0:1:I_NP-1);               % rotating index array (size I_NP)
warm_start_iter = 120;
iter=1;
while iter < warm_start_iter
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    [FM_ui,FM_base,~]=generate_trial(I_strategy,F_weight, F_CR, FM_pop, FVr_bestmemit,I_NP, I_D, FVr_rot);
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    % Boundary Control 
    FM_ui=update (FM_ui,subLbound,subUbound,BRM,FM_base);
    
    %Evaluation of new Pop
    new_pop = minPositionsMatrix;
    new_pop(:,subidx) = FM_ui;
    [S_val_temp, ~]=feval(fnc,new_pop,caseStudyData, otherParameters);
    yhistory = [yhistory; S_val_temp];    
    
    % Elitist Selection
    ind=find(S_val_temp<S_val);
    S_val(ind)=S_val_temp(ind);
    FM_pop(ind,:)=FM_ui(ind,:);
  
    % update best results
    [B,I] = sort(S_val);
    I_best_index = I(1);
    S_bestval = B(1);
    FVr_bestmemit = FM_pop(I_best_index,:); % best member of current iteration
    bestmemit_list = [bestmemit_list; FVr_bestmemit];
    best_x = new_pop(I_best_index,:);
    
    if warm_verbose==1
        fprintf('Fitness value: %f\n',fitMaxVector(gen) );
        fprintf('Generation: %d\n',iter);
    end

    iter = iter +1;
    % record the fitness value
    gen=gen+1;
    fitMaxVector(1,gen) = S_bestval;

end 

%% sande with restart %%
% initialize platform
I_NP = 15;
No_solutions = 15;
otherParameters =setOtherParameters(caseStudyData, No_solutions, 1);
otherParameters.iRuns=iRuns; 
minPositionsMatrix=repmat(FVr_minbound,I_NP,1);
maxPositionsMatrix=repmat(FVr_maxbound,I_NP,1);
deParameters.minPositionsMatrix=minPositionsMatrix;
deParameters.maxPositionsMatrix=maxPositionsMatrix;
subLbound = minPositionsMatrix(:, subidx); 
subUbound = maxPositionsMatrix(:, subidx);

% Trust region adjustment
mean_bestmem = mean(bestmemit_list(end-10:end,:));
thd = (subUbound(1,:) + subLbound(1,:))/2;
up_mask = mean_bestmem > thd;
down_mask = mean_bestmem <= thd;
subLbound_1 = up_mask.*thd + down_mask.*subLbound(1,:);
subLbound = repmat(subLbound_1, I_NP, 1);
subUbound_1 = up_mask.*subUbound(1,:) + down_mask.*thd;
subUbound = repmat(subUbound_1, I_NP, 1);

fprintf('\n******** Main Loop ********\n');
% initial pop
FM_pop2=genpop(I_NP,I_D,subLbound,subUbound);
FM_pop2(1,:) = FM_pop(I(1),:);
FM_pop2(2,:) = FVr_minbound(subidx);

lb_pop = minPositionsMatrix;
lb_pop(:,subidx) =  FM_pop2;

% evaluate initial pop
[S_val, ~]=feval(fnc,lb_pop,caseStudyData,otherParameters);
yhistory = [yhistory; S_val];

[S_bestval,I_best_index] = min(S_val); 
FVr_bestmemit = FM_pop2(I_best_index,:);
best_x = lb_pop(I_best_index,:);

% record the fitness value
gen=gen+1;
fitMaxVector(1,gen) = S_bestval;

% parameter initialization
FEs = 0;
% ccm = deParameters.ccm_init;
ccm = 0.35;
sansde_iter = 182;
linkp = deParameters.link_p;
% linkp = 0.8;
Cycle = 0;
iter = 1;
bestval = S_bestval;
display = 1;
popsize = deParameters.I_NP;

Max_FEs = 182 * I_NP;
old_FEs = Max_FEs;
main_verbose = 1;
while (FEs < Max_FEs)
    Cycle = Cycle + 1;
    dorestart = 1;
    if (FEs + (sansde_iter * popsize) > Max_FEs)
        sansde_iter = ceil((Max_FEs - FEs) / popsize);
    end
    [subpopnew, bestmemnew, bestvalnew, tracerst, ccm, used_FEs, fitMaxVector, gen, yhistory, restart] = sansde_restart(fnc, subidx, FM_pop2, S_val, best_x, bestval, subLbound, subUbound, sansde_iter, ccm, linkp, caseStudyData, otherParameters, fitMaxVector, gen, yhistory, dorestart, main_verbose);
    FEs = FEs + used_FEs;
    FM_pop2 = subpopnew;
    if bestvalnew < bestval
        best_x = bestmemnew;
        FVr_bestmemit = bestmemnew(subidx);
        bestval = bestvalnew;
    end
    iter = iter + 1;

    if(display == 1)
        fprintf(1, 'Cycle = %d, bestval = %e,\n', Cycle, bestval);
    end
    
    if(FEs >= Max_FEs)
        break;
    end
    
    % restart decision making
    if restart ~= 0
        old_FEs = FEs;
        % fprintf('Sande early stop with iter: %d\n',iter)
        break;
    end
end

%% restart 1
remains = 4995 - 35*3*10 - warm_start_iter*10 -I_NP - old_FEs;
if remains > 0 
    fprintf('\n****** Restart Loop1 ******\n');
    FM_pop3=genpop(I_NP,I_D,subLbound,subUbound);
    FM_pop3(1,:) = FVr_bestmemit;
%     FM_pop3(2,:) = FVr_minbound(subidx);
    
    lb_pop = minPositionsMatrix;
    lb_pop(:,subidx) =  FM_pop3;
    
    [S_val, ~]=feval(fnc,lb_pop,caseStudyData,otherParameters);
    yhistory = [yhistory; S_val];
    
    [S_bestval,I_best_index] = min(S_val); % This mean that the best individual correspond to the best worst performance
    FVr_bestmemit = FM_pop3(I_best_index,:); % best member of current iteration
    best_x = lb_pop(I_best_index,:);
    % record the fitness value
    gen=gen+1;
    fitMaxVector(1,gen) = S_bestval;
    
    FEs = 0;
%     ccm = deParameters.ccm_init;
    ccm = 0.35;
    sansde_iter = 140;
    linkp = deParameters.link_p;
%     linkp = 0.8;
    Cycle = 0;
    iter = 1;
    bestval = S_bestval;
    display = 1;
    
    restart_flag = 0;
    Max_FEs = remains -I_NP;
    while (FEs < Max_FEs)
        Cycle = Cycle + 1;
        dorestart = restart;
        if (FEs + (sansde_iter * popsize) > Max_FEs)
            sansde_iter = ceil((Max_FEs - FEs) / popsize);
        end
        [subpopnew, bestmemnew, bestvalnew, tracerst, ccm, used_FEs, fitMaxVector, gen, yhistory,  restart] = sansde_restart(fnc, subidx, FM_pop3, S_val, best_x, bestval, subLbound, subUbound, sansde_iter, ccm, linkp, caseStudyData,otherParameters, fitMaxVector, gen, yhistory, dorestart,main_verbose);
        FEs = FEs + used_FEs;
        FM_pop3 = subpopnew;
        if bestvalnew < bestval
            best_x = bestmemnew;
            FVr_bestmemit = bestmemnew(subidx);
            bestval = bestvalnew;
        end
        iter = iter + 1;
    
        if(display == 1)
            fprintf(1, 'Cycle = %d, bestval = %e,\n', Cycle, bestval);
        end
    
        if(FEs >= Max_FEs)
             break;
        end
    
        if restart==4
            restart_flag = 1;
            remain_FEs = Max_FEs - FEs;
%             fprintf('Sande early stop with iter: %d\n',iter)
%             fprintf('*Finish Restart Loop1---\n');
            break;
        end
    end

    if restart_flag == 1
        %% restart 2
        fprintf('\n****** Restart Loop2 ******\n');
%         FM_pop4=genpop(I_NP,I_D,subLbound,subUbound);
        FM_pop4=genpop_restart(I_NP,I_D,subLbound,subUbound,FVr_bestmemit,4);        
        FM_pop4=update (FM_pop4,subLbound,subUbound,BRM);
        FM_pop4(1,:) = FVr_bestmemit;
        % FM_pop2(2,:) = FM_pop(I(2),:);
        
        lb_pop = minPositionsMatrix;
        lb_pop(:,subidx) =  FM_pop4;
        
        [S_val, ~]=feval(fnc,lb_pop,caseStudyData,otherParameters);
        yhistory = [yhistory; S_val];
        
        [S_bestval,I_best_index] = min(S_val);
        FVr_bestmemit = FM_pop4(I_best_index,:);
        best_x = lb_pop(I_best_index,:);

        % record the fitness value
        gen=gen+1;
        fitMaxVector(1,gen) = S_bestval;    
        
        FEs = 0;
%         ccm = deParameters.ccm_init;
        ccm = 0.35;
        sansde_iter = 80;
        linkp = deParameters.link_p;
%         linkp = 0.8;
        Cycle = 0;
        iter = 1;
        bestval = S_bestval;
        display = 1;
        
        Max_FEs = remain_FEs - I_NP;
        while (FEs < Max_FEs)
            Cycle = Cycle + 1;
            dorestart = 0;        
            if (FEs + (sansde_iter * popsize) > Max_FEs)
                sansde_iter = ceil((Max_FEs - FEs) / popsize);
            end
            [subpopnew, bestmemnew, bestvalnew, tracerst, ccm, used_FEs, fitMaxVector, gen, yhistory, restart] = sansde_restart(fnc, subidx, FM_pop4, S_val, best_x, bestval, subLbound, subUbound, sansde_iter, ccm, linkp, caseStudyData,otherParameters, fitMaxVector, gen, yhistory, dorestart, main_verbose);
            FEs = FEs + used_FEs;
            FM_pop4 = subpopnew;
            if bestvalnew < bestval
                best_x = bestmemnew;
                FVr_bestmemit = bestmemnew(subidx);
                bestval = bestvalnew;
            end
            iter = iter + 1;   
        
            if(display == 1)
                fprintf(1, 'Cycle = %d, bestval = %e,\n', Cycle, bestval);
            end
        
            if(FEs >= Max_FEs)
                 break;
            end
    
        end
    end
%     fprintf('---Finish Restart Loop2---\n');
end
fprintf('---------- Finish ReSaDE ----------\n');

fprintf('==================== Finish trial %d =================\n', iRuns);
fprintf('\n')
Fit_and_p=[fitMaxVector(I_itermax) 0]; %;p2;p3;p4]

 
% VECTORIZED THE CODE INSTEAD OF USING FOR
function pop=genpop(a,b,lowMatrix,upMatrix)
pop=unifrnd(lowMatrix,upMatrix,a,b);

% VECTORIZED THE CODE INSTEAD OF USING FOR
function pop=genpop_restart(a, b, lowMatrix, upMatrix, basepop, n)
pop=unifrnd(lowMatrix, upMatrix, a, b);
range = upMatrix(1,:) - lowMatrix(1,:);
upmu = lowMatrix(1,:) + range/4*3;
lowmu = lowMatrix(1,:) + range/4;
N = length(range);
mu = lowmu;
for i=1:N
    if basepop(i) < lowMatrix(i) + range(i)/2
        mu(i) = upmu(i);
    end
end

sigma = range/2;
for i = 0:n
    pop((end-i),:) = mu + sigma.*randn(1,b);
end

% VECTORIZED THE CODE INSTEAD OF USING FOR
function p=update(p,lowMatrix,upMatrix,BRM,FM_base)
switch BRM
    case 1 %Our method
        %[popsize,dim]=size(p);
        [idx] = find(p<lowMatrix);
        p(idx)=lowMatrix(idx);
        [idx] = find(p>upMatrix);
        p(idx)=upMatrix(idx);
    case 2 %Random reinitialization
        [idx] = [find(p<lowMatrix);find(p>upMatrix)];
        replace=unifrnd(lowMatrix(idx),upMatrix(idx),length(idx),1);
        p(idx)=replace;
    case 3 %Bounce Back
      [idx] = find(p<lowMatrix);
      p(idx)=unifrnd(lowMatrix(idx),FM_base(idx),length(idx),1);
        [idx] = find(p>upMatrix);
      p(idx)=unifrnd(FM_base(idx), upMatrix(idx),length(idx),1);
end


function [FM_ui,FM_base,msg]=generate_trial(method,F_weight, F_CR, FM_pop, FVr_bestmemit,I_NP,I_D,FVr_rot)

  FM_popold = FM_pop;                  % save the old population
  FVr_ind = randperm(4);               % index pointer array
  FVr_a1  = randperm(I_NP);                   % shuffle locations of vectors
  FVr_rt  = rem(FVr_rot+FVr_ind(1),I_NP);     % rotate indices by ind(1) positions
  FVr_a2  = FVr_a1(FVr_rt+1);                 % rotate vector locations
  FVr_rt  = rem(FVr_rot+FVr_ind(2),I_NP);
  FVr_a3  = FVr_a2(FVr_rt+1);   
  FVr_rt  = rem(FVr_rot+FVr_ind(3),I_NP);
  FVr_a4  = FVr_a3(FVr_rt+1); 
  FVr_rt  = rem(FVr_rot+FVr_ind(4),I_NP);
  FVr_a5  = FVr_a4(FVr_rt+1); 
  FM_pm1 = FM_popold(FVr_a1,:);             % shuffled population 1
  FM_pm2 = FM_popold(FVr_a2,:);             % shuffled population 2
  FM_pm3 = FM_popold(FVr_a3,:);             % shuffled population 3
  FM_pm4 = FM_popold(FVr_a4,:);             % shuffled population 4
  FM_pm5 = FM_popold(FVr_a5,:);             % shuffled population 5  
  FM_mui = rand(I_NP,I_D) < F_CR;  % all random numbers < F_CR are 1, 0 otherwise
  FM_mpo = FM_mui < 0.5;    % inverse mask to FM_mui

    switch method
        case 1
            FM_ui = FM_pm3 + F_weight*(FM_pm1 - FM_pm2);   % differential variation
            FM_ui = FM_popold.*FM_mpo + FM_ui.*FM_mui;     % crossover
            FM_base = FM_pm3;
            msg=' DE/rand/bin';
        case 2
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%  
            %VEC by FLC
            FM_bm=repmat(FVr_bestmemit,I_NP,1);
            FM_ui = FM_popold + F_weight*(FM_bm-FM_popold) + F_weight*(FM_pm1 - FM_pm2);
            FM_ui = FM_popold.*FM_mpo + FM_ui.*FM_mui;
            FM_base = FM_bm;
            msg=' DE/current-to-best/1';
        case 3
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%  
            %VEC by FLC
            FM_ui = FM_pm5 + F_weight*(FM_pm1 - FM_pm2) + F_weight*(FM_pm3 - FM_pm4); 
            FM_ui = FM_popold.*FM_mpo + FM_ui.*FM_mui;
            FM_base = FM_pm5;
            msg=' DE/rand/2';    
        case 4
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%  
            %VEC by FLC
            FM_bm=repmat(FVr_bestmemit,I_NP,1);
            FM_ui = FM_bm + F_weight*(FM_pm1 - FM_pm2) + F_weight*(FM_pm3 - FM_pm4); 
            FM_ui = FM_popold.*FM_mpo + FM_ui.*FM_mui;
            FM_base = FM_bm;
            msg=' DE/best/2';               
    end
return
